#!/usr/bin/perl -w

use strict;

sub reduce()
{
  my $source=$_[0];
  my @cycle=split(/\s/, $_[1]);

  my %evts=();
  my $e=0;
  while($source =~ /^(\S+)\s*\(([^\)]*) thread (\d+)\)\s*(.+)?$/)
  {
    $evts{$e}=
    {
      obj => $1,
      loc => $2,
      thread => $3,
      rm => 0
    };
    $e++;
    defined($4) or last;
    $source=$4;
  }

  my $fixedpoint=0;
  while(!$fixedpoint && scalar(@cycle)>2)
  {
    $fixedpoint=1;
    my $i=0;
    my $e_i=-1;
    foreach(sort { $a <=> $b } keys %evts)
    {
      next unless($evts{$_}{obj} =~ /^c::/ && !$evts{$_}{rm});
      $e_i=$_;
      last;
    }
    ($e_i >= 0) or die "No shared memory access\n";

    while($fixedpoint && $i<scalar(@cycle))
    {
      my $i_p=($i+1) % scalar(@cycle);

      my $cur=$cycle[$i];
      my $next=$cycle[$i_p];

      my $e_cur=($e_i+1) % scalar(keys %evts);
      while($e_cur!=$e_i &&
            (!($evts{$e_cur}{obj} =~ /^c::/) || $evts{$e_cur}{rm}))
      {
        $e_cur=($e_cur+1) % scalar(keys %evts);
      }

      my $e_next=($e_cur+1) % scalar(keys %evts);
      while($e_next!=$e_i &&
            (!($evts{$e_next}{obj} =~ /^c::/) || $evts{$e_next}{rm}))
      {
        $e_next=($e_next+1) % scalar(keys %evts);
      }

      ## print "cur=$cur\n";
      ## print "next=$next\n";
      ## print "i=$i\n";
      ## print "i_p=$i_p\n";
      ## print "e_i=$e_i\n";
      ## print "e_cur=$e_cur\n";
      ## print "e_next=$e_next\n";

      if("$cur $next" =~ /^(Fr|Ws)(i|e) Ws(i|e)$/)
      {
        my $e=($evts{$e_i}{thread}!=$evts{$e_next}{thread}) ? "e" : "i";
        $cycle[$i]="$1$e";
        splice(@cycle, $i_p, 1);
        $evts{$e_cur}{rm}=1;
        $fixedpoint=0;
      }
      elsif("$cur $next" =~ /^Rf(i|e) Fr(i|e)$/)
      {
        my $e=($evts{$e_i}{thread}!=$evts{$e_next}{thread}) ? "e" : "i";
        $cycle[$i]="Ws$e";
        splice(@cycle, $i_p, 1);
        $evts{$e_cur}{rm}=1;
        $fixedpoint=0;
      }
      elsif("$cur $next" =~ /^(Pos..|DpDatas.|Rfi|Fri|Wsi) (Po|Sync|LwSync)(s|d).(.)$/)
      {
        $cycle[$i]="$2$3D$4";
        my $d=$1;
        $d=($d =~ /^(PosW.|Rfi|Wsi)/) ? "W" : "R";
        $cycle[$i]=~ s/D/$d/;
        splice(@cycle, $i_p, 1);
        $evts{$e_cur}{rm}=1;
        $fixedpoint=0;
      }
      elsif("$cur $next" =~ /^(Pos..|DpDatas.|Rfi|Fri|Wsi) (DpData)(s|d)(.)$/)
      {
        $cycle[$i]="$3D$4";
        my $d=$1;
        $d=($d =~ /^(PosW.|Rfi|Wsi)/) ? "W" : "R";
        $cycle[$i]=~ s/D/$d/;
        if($d =~ /^DpDatas/)
        {
          $cycle[$i]=~ s/^/DpData/;
        }
        else
        {
          $cycle[$i]=~ s/^/Po/;
        }
        splice(@cycle, $i_p, 1);
        $evts{$e_cur}{rm}=1;
        $fixedpoint=0;
      }
      elsif("$cur $next" =~ /^(Pod..|DpDatad.) (Po|Sync|LwSync)(s|d).(.)$/)
      {
        my $s=($evts{$e_i}{obj} ne $evts{$e_next}{obj}) ? "d" : "s";
        $cycle[$i]="$2${s}D$4";
        my $d=$1;
        $d=($d =~ /^PodW./) ? "W" : "R";
        $cycle[$i]=~ s/D/$d/;
        splice(@cycle, $i_p, 1);
        $evts{$e_cur}{rm}=1;
        $fixedpoint=0;
      }

      $i++;
      $e_i=($e_i+1) % scalar(keys %evts);
      while(!($evts{$e_i}{obj} =~ /^c::/) || $evts{$e_i}{rm})
      {
        $e_i=($e_i+1) % scalar(keys %evts);
      }
    }
  }

  my @ret=();
  push @ret, "";
  push @ret, join(" ", @cycle);

  my %thread_map=();
  foreach(keys %evts)
  {
    next if($evts{$_}{rm});
    $thread_map{$evts{$_}{thread}}=0;
  }
  my $i=0;
  foreach(sort { $a <=> $b } keys %thread_map)
  {
    $thread_map{$_}=$i;
    $i++;
  }

  foreach(sort { $a <=> $b } keys %evts)
  {
    next if($evts{$_}{rm});
    $ret[0].=" " if($ret[0] ne "");

    $ret[0].="$evts{$_}{obj} ($evts{$_}{loc} thread $thread_map{$evts{$_}{thread}})";
  }

  return @ret;
}

sub is_critical()
{
  my @cycle=split(/\s/, $_[0]);
  my $ret=1;

  my $first=undef;
  my $prev_was_pod=0;
  my $consecutive_com=0;

  while($ret && scalar(@cycle))
  {
    my $front=shift @cycle;
    $first=$front if(!defined($first));

    $ret=($front =~ /^(Pos..|DpDatas.|Rfi|Wsi|Fri|Syncs..|LwSyncs..)$/) ? 0 : 1;

    if($front =~ /^(Pod..|DpDatad.|Syncd..|LwSyncd..)$/)
    {
      $consecutive_com=0;
      if($prev_was_pod ||
         (!scalar(@cycle) && $first =~ /^(Pod..|DpDatad.|Syncd..|LwSyncd..)$/))
      {
        $ret=0;
      }

      $prev_was_pod=1;
    }
    elsif($consecutive_com == 2 ||
          ($consecutive_com == 1 && !scalar(@cycle) && $first =~ /^(Rfe|Wse|Fre)$/))
    {
      $ret=0;
    }
    else
    {
      $prev_was_pod=0;
      $consecutive_com++;
    }
  }

  return $ret;
}

sub is_valid_uniproc()
{
  my @cycle=split(/\s/, $_[0]);
  my $ret=1;

  return 0 if(scalar(@cycle)<2);

  if(scalar(@cycle)==2)
  {
    if($_[0] =~ /^Ws(e|i) Ws(e|i)$/ ||
       $_[0] eq "Fre Rfe" ||
       $_[0] eq "Rfe Fre")
    {
      return 0;
    }
  }

  if($_[0] =~ /^(((Po(s|d)(W|R){2})|(DpData(s|d)(W|R))|((Lw)?Sync(s|d)(W|R){2})) )+((Po(s|d)(W|R){2})|(DpData(s|d)(W|R))|((Lw)?Sync(s|d)(W|R){2}))$/)
  {
    return 0;
  }

  while($ret && scalar(@cycle))
  {
    my $front=shift @cycle;

    $ret=($front =~ /^(Pos..|DpDatas.|Rfi|Wsi|Fri|Syncs..|LwSyncs..)$/ ||
          $front =~ /^(Rfe|Wse|Fre)$/) ? 1 : 0;
  }

  return $ret;
}

while(<>)
{
  chomp;
  (/^([^;]+);([^;]+)$/) or die "Invalid input $_\n";

  my $source=$1;
  my $named=$2;

  my ($source_red,$cycle)=&reduce($source, $named);

  if(&is_critical($cycle))
  {
    print "$source;$named;$source_red;$cycle;critical\n";
  }
  elsif(&is_valid_uniproc($cycle))
  {
    print "$source;$named;$source_red;$cycle;uniproc\n";
  }
  ## else
  ## {
  ##   print "not a critical cycle or invalid: $cycle\n";
  ## }
}

